@@ -5,7 +5,7 @@ use 5.006;
 use strict;
 use warnings;
 use Carp;
-our $VERSION = '0.03_1';
+our $VERSION = '0.05';
 my %types = (
     go		=> 1,
     othello	=> 2,
@@ -49,7 +49,7 @@ values (* marks defaults):
     game       *go, othello, renju, gomoku
     white      Any text, default: "Miss White"
     black      Any text, default: "Mr Black"
-	skip_i     Truth value; whether 'i' should be skipped; true by default
+	skip_i     Truth value; whether 'i' should be skipped; false by default
     referee    Any subroutine, default: sub {1} # (All moves are valid) 
 
 The referee subroutine takes a board object and a piece object, and
@@ -71,18 +71,19 @@ sub new {
     
     my $board = bless {
         game => $game,
-        board => [ [ (undef) x $size ] x $size ],
         moves => [],
         size => $size,
         black => $opts{black} || "Mr. Black",
         white => $opts{white} || "Miss White",
-        callbacks => [],
+        callbacks => {},
         referee => $opts{referee} || sub { 1 },
         move => 1,
         magiccookie => "a0000",
         turn => 'b',
+		skip_i => (defined $opts{skip_i} ? $opts{skip_i} : 0)
     }, $class;
 
+	for (0 .. ($size-1)) { push @{$board->{board}}, [ (undef) x $size ]; }
 	$board->{hoshi} = $board->_calc_hoshi;
 
 	return $board;
@@ -103,21 +104,45 @@ true.
 
 sub move {
     my ($self, $move) = @_;
+
 	my ($x,$y) = $self->_pos2grid($move, $self->skip_i);
 
-    $move = _check_pos($self,$move);
+    $self->_check_pos($move);
     my $stat = $self->{referee}->($self,$move);
 
     return $stat if !$stat;
     $self->{board}[$x][$y] = bless {
-        colour => $self->{turn},
-        move  => $self->{move},
-        position => $move,
-        board => $self
+        colour   => $self->{turn},
+        move     => $self->{move},
+		xy       => [$x, $y],
+        board    => $self
     }, "Games::Goban::Piece";
-    push @{$self->{moves}}, $self->{board}[$x][$y];
+    push @{$self->{moves}}, { 
+		player => $self->{turn},
+		piece => $self->{board}[$x][$y]
+	};
     $self->{move}++;
     $self->{turn} = $self->{turn} eq "b" ? "w" : "b";
+
+	while (my ($key, $cb) = each %{$self->{callbacks}}) { $cb->($key,$self) }
+}
+
+=head2 pass
+
+This method causes the current player to pass.  At present, nothing happens for
+two subsequent passes.
+
+=cut
+
+sub pass {
+	my $self = shift;
+
+	push @{$self->{moves}}, {
+		player => $self->{turn},
+		piece => undef
+	};
+	$self->{move}++;
+    $self->{turn} = $self->{turn} eq "b" ? "w" : "b";
 }
 
 =head2 get
@@ -126,8 +151,9 @@ sub move {
 
 Gets the C<Games::Goban::Piece> object at the given location, if there
 is one. Locations are specified as per SGF - a 19x19 board starts from
-C<aa> in the top left corner, with C<tt> in the bottom right. C<i> does
-not exist.
+C<aa> in the top left corner, with C<ss> in the bottom right.  (If the skip_i
+option was set while creating the board, C<tt> is the bottom right and there
+are no C<i> positions.  This allows for traditional notation.)
 
 =cut
 
@@ -160,7 +186,7 @@ Returns a list of hoshi points.
 sub hoshi {
 	my $self = shift;
 
-	map { $self->_grid2pos( @$_ ) } @{$self->{hoshi}};
+	map { $self->_grid2pos( @$_, $self->skip_i ) } @{$self->{hoshi}};
 }
 
 =head2 is_hoshi
@@ -191,7 +217,8 @@ sub as_sgf {
 
 	$sgf .= "(;GM[$types{$self->{game}}]FF[4]AP[Games::Goban]SZ[$self->{size}]PB[$self->{black}]PW[$self->{white}]\n";
 	foreach (@{$self->{moves}}) {
-		$sgf .= ";" . uc($_->color) ."[".$_->position."]";
+		$sgf .= ";" . uc($_->{player}) .  "[".  ($_->{piece} ?
+		$self->_grid2pos(@{$_->{piece}->_xy},0) : '') . "]";
 	}
 	$sgf .= ")\n";
 	
@@ -204,8 +231,8 @@ sub as_sgf {
 
 Returns a printable text picture of the board, similar to that printed
 by C<gnugo>. Black pieces are represented by C<X>, white pieces by C<O>,
-and the latest move is bracketed. I<hoshi> points are in their normal
-position for Go, and printed as an C<+>. Coordinates are not printed by
+and the latest move is enclosed in parentheses. I<hoshi> points are in their
+normal position for Go, and printed as an C<+>. Coordinates are not printed by
 default, but can be enabled as suggested in the synopsis.
 
 =cut
@@ -215,9 +242,8 @@ sub as_text {
     my %opts = @_;
 	my @hoshi = $board->hoshi;
     my $text;
-    for my $y (0 .. ($board->size - 1)) {
-        $text .= sprintf("%2i: ", $board->size - (ord($y) - ORIGIN)) 
-            if $opts{coords};
+    for (my $y = $board->size - 1; $y >= 0; $y--) {
+        $text .= substr($board->_grid2pos(0,$y,$board->skip_i),1,1) . ': ' if $opts{coords};
 		for my $x (0 .. ($board->size - 1)) {
 			my $pos = $board->_grid2pos($x,$y,$board->skip_i);
             my $p = $board->get($pos);
@@ -229,6 +255,11 @@ sub as_text {
         }
         $text .= "\n";
     }
+	if ($opts{coords}) {
+		$text .= '   ';
+		for (0 .. ($board->size - 1)) { $text .= substr($board->_grid2pos($_,0,$board->skip_i),0,1) . ' '; }
+        $text .= "\n";
+	}
     return $text;
 }
 
@@ -236,16 +267,18 @@ sub as_text {
 
     my $key = $board->register(\&callback);
 
-Register a calllback to be called after every move is made. This is
-useful for analysis programs which wish to maintain statistics on the
-board state. The C<key> returned from this can be fed to...
+Register a callback to be called after every move is made. This is useful for
+analysis programs which wish to maintain statistics on the board state. The
+C<key> returned from this can be fed to...
 
 =cut
 
 sub register {
     my ($board, $cb) = @_;
-    push @{$board->{callbacks}}, $cb;
-    return ++$board->{magiccookie};
+	my $key = ++$board->{magiccookie};
+	$board->{callbacks}{$key} = $cb;
+	$board->{notes}->{$key} = {};
+	return $key;
 }
 
 =head2 notes
@@ -295,9 +328,9 @@ producing SGF output.)
 sub skip_i { return (shift)->{skip_i} }
 
 # This method accepts a position string and checks whether it is a valid
-# position on the given board.  If it is, the string is returned.  Otherwise,
-# it carps that the position is not on the board.  It does this by calling
-# _check_grid, also below.
+# position on the given board.  If it is, 1 is returned.  Otherwise, it carps
+# that the position is not on the board.  It does this by calling _check_grid,
+# also below.
 
 sub _check_pos {
     my $self = shift;
@@ -305,18 +338,17 @@ sub _check_pos {
 	
 	my ($x, $y) = $self->_pos2grid($pos,$self->skip_i);
 
-	return $pos if $self->_check_grid($x,$y);
+	return $self->_check_grid($x,$y);
 }
 
 sub _check_grid {
 	my $self = shift;
 	my ($x, $y) = @_;
 
-	return ($x,$y) if 
+	return 1 if 
 		(($x < $self->size) and ($y < $self->size));
 	
-	local $Carp::CarpLevel=2;
-	croak "position '" . $self->_pos2grid($x,$y,$self->skip_i) . "' not on board";
+	croak "position '" . $self->_grid2pos($x,$y,$self->skip_i) . "' not on board";
 }
 
 # This method returns a list of the hoshi points that should be found on the
@@ -325,7 +357,7 @@ sub _check_grid {
 sub _calc_hoshi {
 	my $self = shift;
 	my $size = $self->size;
-	my $half = ($size + 1) / 2;
+	my $half = ($size - 1) / 2;
 
 	my @hoshi = ();
 
@@ -336,18 +368,18 @@ sub _calc_hoshi {
 	return \@hoshi unless $margin;
 	
 	push @hoshi, (
-		[ $margin,         $margin ],          # top left
-		[ $size-$margin+1, $margin  ],         # top right
-		[ $margin,         $size-$margin+1  ], # bottom left
-		[ $size-$margin+1, $size-$margin+1  ]  # bottom right
+		[ $margin-1,     $margin-1     ], # top left
+		[ $size-$margin, $margin-1     ], # top right
+		[ $margin-1,     $size-$margin ], # bottom left
+		[ $size-$margin, $size-$margin ]  # bottom right
 	);
 
 	if (($size % 2) && ($size > 9)) {
 		push @hoshi, (
-			[ $half,           $margin ],        # top center
-			[ $margin,         $half ],          # middle left
-			[ $size-$margin+1, $half ],          # middle right
-			[ $half,           $size-$margin+1 ] # bottom center
+			[ $half,         $margin-1 ],    # top center
+			[ $margin-1,     $half ],        # middle left
+			[ $size-$margin, $half ],        # middle right
+			[ $half,         $size-$margin ] # bottom center
 		);
 	}
 
@@ -377,11 +409,11 @@ sub _grid2pos {
 
 	if ($skip_i) {
 		for ($x,$y) {
-			$_++ if ($_ > 8);
+			$_++ if ($_ >= 8);
 		}
 	}
 
-	return chr(ORIGIN - 1 + $x) . chr(ORIGIN - 1 + $y);
+	return chr(ORIGIN + $x) . chr(ORIGIN + $y);
 }
 
 # This method accepts an 'xy' position string and returns the (x,y) indexes
@@ -443,7 +475,13 @@ one in that position.
 
 =cut
 
-sub position { $_[0]->{position} }
+sub position {
+	my $piece = shift;
+
+	$piece->board->_grid2pos(@{$piece->_xy},$piece->board->skip_i);
+}
+
+sub _xy { $_[0]->{xy} }
 
 =head1 move
 
@@ -463,6 +501,28 @@ sub board { $_[0]->{board} }
 
 1;
 
+=head1 TODO
+
+=over
+
+=item * 
+
+add C<<$board->pass>>
+
+=item *
+
+possibly enable C<<$board->move('')>> to pass
+
+=item *
+
+produce example referee
+
+=item *
+
+produce sample method for removing captured stones
+
+=back
+
 =head1 SEE ALSO
 
 Smart Game Format: http://www.red-bean.com/sgf/
